// graphics_impl.cpp
// Revision 21-jan-2010

#include "graphics_impl.h"


#include "error.h"
#include "trace.h"
#include "charset.h"
#include "util.h"

#include <iostream>
#include <memory>

#include <string.h>
#include <math.h>


using std::cerr;
using std::endl;
using std::swap;
using std::auto_ptr;

using util::to_string;
using util::dim_array;


namespace blassic {


charset::chardataset charset::data;

const charset::chardataset * charset::default_charset=
	& charset::default_data;


#ifdef BLASSIC_HAS_GRAPHICS


namespace graphics_impl {


QueueKey::QueueKey ()
{
}

QueueKey::~QueueKey ()
{
}

void QueueKey::push (const std::string & str)
{
	lock ();
	try {
		q.push (str);
	}
	catch (...)
	{
		unlock ();
		throw;
	}
	unlock ();
}

std::string QueueKey::pop ()
{
	lock ();
	std::string str= q.front ();
	q.pop ();
	unlock ();
	return str;
}

bool QueueKey::empty ()
{
	lock ();
	bool r= q.empty ();
	unlock ();
	return r;
}

void QueueKey::erase ()
{
	lock ();
	while (! q.empty () )
		q.pop ();
	unlock ();
}

void QueueKey::lock () { }

void QueueKey::unlock () { }


Image::Image ()
{
}

Image::~Image ()
{
}


Painter::Visited::Visited (int width_n, int height_n) :
	v (width_n * height_n),
	width (width_n)
{
}

bool Painter::Visited::check (int x, int y) const
{
	return v [x + y * width];
}

void Painter::Visited::set (int x, int y)
{
	v [x + y * width]= '\1';
}


Painter::Painter (int minx, int miny, int maxx, int maxy) :
	//visited (maxx - minx + 1, v_visited_t (maxy - miny + 1) ),
	visited (maxx - minx + 1, maxy - miny + 1),
	min_x (minx),
	min_y (miny),
	max_x (maxx),
	max_y (maxy)
{
	TRFDEBS (min_x << ", " << min_y << ", " << max_x << ", " << max_y);
}

Painter::~Painter ()
{
	TRF;
}

bool Painter::check (Point p)
{
	int x= p.x, y= p.y;
	if (x < min_x || y < min_y || x > max_x || y > max_y)
		return false; 
	//v_visited_t::reference v= visited [x - min_x] [y - min_y];
	//if (v)
	//	return false;

	if (visited.check (x - min_x, y - min_y) )
		return false;

	//v= true;
	visited.set (x - min_x, y - min_y);
	return do_check (p);
}

void Painter::gf (Point p)
{
	if (! check (p) )
		return;
	plot (p);
	nextlist.push_back (Point (p.x - 1, p.y) );
	nextlist.push_back (Point (p.x + 1, p.y) );
	nextlist.push_back (Point (p.x, p.y - 1) );
	nextlist.push_back (Point (p.x, p.y + 1) );
}

void Painter::paint (int x, int y)
{
	TRF;

	processlist.push_back (Point (x, y) );
	while (! processlist.empty () )
	{
		nextlist.clear ();
		while (! processlist.empty () )
		{
			Point cp= processlist.back ();
			processlist.pop_back ();
			gf (cp);
		}
		processlist= nextlist;
	}
}


GraphicsSystem::GraphicsSystem ()
{
	TRF;
}

GraphicsSystem::~GraphicsSystem ()
{
}

void GraphicsSystem::check_events ()
{
	// Do nothing
}


Screen::Screen (GraphicsSystem & system_n) :
	system (system_n),
	active (false),
	scwidth (0),
	scheight (0),
	realwidth (0),
	realheight (0),
	activetransform (TrIdentity),
	//colors_inited (false),
	transparent_text (false),
	limited (false),
	rotatemode (RotateNone),
	numsynchrosaved (0),
	synchrosaved (false),
	maskvalue (maskvaluedefault),
	maskdrawfirst (true),
	maskpos (0),
	originx (0),
	originy (0)
{
	TRF;
}

Screen::~Screen ()
{
	TRF;
}

GraphicsSystem & Screen::get_system ()
{
	return system;
}

bool Screen::is_active () const
{
	return active;
}

void Screen::create (int width_n, int height_n, RotateMode rotmode,
	Inkset inkset_n, InkNum pen_n, InkNum paper_n,
	ZoomValue charx, ZoomValue chary)
{
	TRF;

	inkset= inkset_n;
	scwidth= width_n;
	scheight= height_n;
	realwidth= scwidth;
	realheight= scheight;

	zoomtextx= charx;
	zoomtexty= chary;

	ASSERT (zoomtextx != 0);
	ASSERT (zoomtexty != 0);

	charwidth= 8 * zoomtextx;
	charheight= 8 * zoomtexty;

	//maxtcol= screenwidth / charwidth;
	//maxtrow= screenheight / charheight;
	maxtcol= scwidth / charwidth;
	maxtrow= scheight / charheight;

	lastx= 0; lasty= 0;

	activetransform= TrIdentity;

	default_pen= pen_n;
	default_paper= paper_n;
	graphics_pen= pen_n;
	graphics_paper= paper_n;

	transparent_text= false;

	//limited= false;
	set_limits (0, scwidth, 0, scheight);

	switch (rotmode)
	{
	case RotateNone:
		// Nothing to do.
		break;
	case Rotate90:
		swap (realwidth, realheight);
		break;
	default:
		throw ErrBlassicInternal;
	}
	rotatemode= rotmode;

	numsynchrosaved= 0;

	maskvalue= maskvaluedefault;
	maskdrawfirst= true;
	maskpos= 0;

	originx= 0;
	originy= 0;

	raw_create (scwidth, scheight, realwidth, realheight,
		rotmode, paper_n);

	active= true;

	show ();
}

void Screen::flush ()
{
}

void Screen::suspend_synchro ()
{
	if (++numsynchrosaved > 1)
	{
		if (showdebuginfo () )
			cerr << "synchronize_suspend called several times" <<
				endl;
	}
	else
	{
		synchrosaved= get_synchro ();
		set_synchro (false);
	}
}

void Screen::restart_synchro ()
{
	if (numsynchrosaved == 0)
	{
		if (showdebuginfo () )
			cerr << "uexpected call to synchronize_restart" <<
				endl;
	}
	else
	{
		if (--numsynchrosaved == 0)
			set_synchro (synchrosaved);
	}
}

int Screen::get_width () const
{
	return scwidth;
}

int Screen::get_height () const
{
	return scheight;
}

void Screen::set_transform (TransformType newtransform)
{
	activetransform= newtransform;
}


const int LancelotsFavouriteColour= 0x0204FB;
// http://mindprod.com/unmainnaming.html


void Screen::setcpcink (InkNum inknum, int cpccolor)
{
	// These rgb values are taken from screen captures
	// of the WinAPE2 Amstrad CPC emulator.
	static const ColorRGB cpctable []= {
		ColorRGB (   0,   0,   0 ), // Black
		ColorRGB (   0,   0,  96 ), // Blue
		ColorRGB (   0,   0, 255 ), // Bright blue
		ColorRGB (  96,   0,   0 ), // Red
		ColorRGB (  96,   0,  96 ), // Magenta
		ColorRGB (  96,   0, 255 ), // Mauve
		ColorRGB ( 255,   0,   0 ), // Bright red
		ColorRGB ( 255,   0,  96 ), // Purple
		ColorRGB ( 255,   0, 255 ), // Bright magenta
		ColorRGB (   0, 103,   0 ), // Green
		ColorRGB (   0, 103,  96 ), // Cyan
		ColorRGB (   0, 103, 255 ), // Sky blue
		ColorRGB (  96, 103,   0 ), // Yellow
		ColorRGB (  96, 103,  96 ), // White
		ColorRGB (  96, 103, 255 ), // Pastel blue
		ColorRGB ( 255, 103,   0 ), // Orange
		ColorRGB ( 255, 103,  96 ), // Pink
		ColorRGB ( 255, 103, 255 ), // Pastel magenta
		ColorRGB (   0, 255,   0 ), // Bright green
		ColorRGB (   0, 255,  96 ), // Sea green
		ColorRGB (   0, 255, 255 ), // Bright cyan
		ColorRGB (  96, 255,   0 ), // Lime green
		ColorRGB (  96, 255,  96 ), // Pastel green
		ColorRGB (  96, 255, 255 ), // Pastel cyan
		ColorRGB ( 255, 255,   0 ), // Bright yellow
		ColorRGB ( 255, 255,  96 ), // Pastel yellow
		ColorRGB ( 255, 255, 255 ), // Brigth white
	};
	ASSERT (cpccolor >= 0 &&
		cpccolor < static_cast <int> (dim_array (cpctable) ) );
	const ColorRGB & rgb= cpctable [cpccolor];

	setink (inknum, rgb);
}

void Screen::set_standard_default_inks ()
{
	static const ColorRGB assignRGB []= {
		ColorRGB (    0,    0,    0 ),
		ColorRGB (    0,    0, 0xA8 ),
		ColorRGB (    0, 0xA8,    0 ),
		ColorRGB (    0, 0xA8, 0xA8 ),
		ColorRGB ( 0xA8,    0,    0 ),
		ColorRGB ( 0xA8,    0, 0xA8 ),
		ColorRGB ( 0xA8, 0x54,    0 ),
		ColorRGB ( 0xA8, 0xA8, 0xA8 ),

		ColorRGB ( 0x54, 0x54, 0x54 ),
		ColorRGB ( 0x54, 0x54, 0xFF ),
		ColorRGB ( 0x54, 0xFF, 0x54 ),
		ColorRGB ( 0x54, 0xFF, 0xFF ),
		ColorRGB ( 0xFF, 0x54, 0x54 ),
		ColorRGB ( 0xFF, 0x54, 0xFF ),
		ColorRGB ( 0xFF, 0xFF, 0x54 ),
		ColorRGB ( 0xFF, 0xFF, 0xFF )
	};
	ASSERT (dim_array (assignRGB) == PREDEFINED_INKS);

	for (InkNum i= 0; i < PREDEFINED_INKS; ++i)
	{
		const ColorRGB & rgb= assignRGB [i];
		setink (i, rgb);
	}
}

void Screen::set_cpc_default_inks ()
{
	TRF;

	static const InkNum default_ink []=
		{ 1, 24, 20, 6, 26, 0, 2, 8, 10, 12, 14, 16, 18, 22, 1, 16 };
	// The last two are blinking on the CPC, we use the first color.
	ASSERT (dim_array (default_ink) == PREDEFINED_INKS);

	for (InkNum i= 0; i < PREDEFINED_INKS; ++i)
		setcpcink (i, default_ink [i] );
}

void Screen::set_spectrum_default_inks ()
{
	TRF;

	// Taken from a screen capture of the Spectaculator Spectrum Emulator.
	static const ColorRGB spectrumtable []=
	{
		ColorRGB (   0,   0,   0 ), // Black
		ColorRGB (   0,   0, 207 ), // Blue
		ColorRGB ( 207,   0,   0 ), // Red
		ColorRGB ( 207,   0, 207 ), // Magenta
		ColorRGB (   0, 200,   0 ), // Green
		ColorRGB (   0, 200, 207 ), // Cyan
		ColorRGB ( 207, 200,   0 ), // Yellow
		ColorRGB ( 207, 200, 207 ), // White
		ColorRGB (   0,   0,   0 ), // Black bright
		ColorRGB (   0,   0, 255 ), // Blue bright
		ColorRGB ( 255,   0,   0 ), // Red bright
		ColorRGB ( 255,   0, 255 ), // Magenta bright
		ColorRGB (   0, 248,   0 ), // Green bright
		ColorRGB (   0, 248, 255 ), // Cyan bright
		ColorRGB ( 255, 248,   0 ), // Yellow bright
		ColorRGB ( 255, 248, 255 ), // White bright
	};
	ASSERT (dim_array (spectrumtable) == PREDEFINED_INKS);

	for (InkNum i= 0; i < PREDEFINED_INKS; ++i)
		setink (i, spectrumtable [i] );
}

void Screen::set_msx_default_inks ()
{
	TRF;

	// Values by courtesy of abnog
	static const ColorRGB msxtable []=
	{
		ColorRGB (   0,   0,   0 ), // Transparent, not used in Blassic
		ColorRGB (   0,   0,   0 ), // Black
		ColorRGB (  33, 200,  66 ), // Medium green
		ColorRGB (  94, 220, 129 ), // Light green
		ColorRGB (  84,  85, 237 ), // Dark blue
		ColorRGB ( 125, 118, 252 ), // Light blue
		ColorRGB ( 212,  82,  77 ), // Dark red
		ColorRGB (  66, 235, 245 ), // Cyan
		ColorRGB ( 252,  85,  84 ), // Medium red
		ColorRGB ( 235, 121, 120 ), // Ligh red
		ColorRGB ( 212, 193,  84 ), // Dark yellow
		ColorRGB ( 230, 206, 128 ), // Light yellow
		ColorRGB (  33, 176,  59 ), // Dark green
		ColorRGB ( 201,  91, 186 ), // Magenta
		ColorRGB ( 204, 204, 204 ), // Gray
		ColorRGB ( 255, 255, 255 ), // White
	};
	ASSERT (dim_array (msxtable) == PREDEFINED_INKS);

	for (InkNum i= 0; i < PREDEFINED_INKS; ++i)
		setink (i, msxtable [i] );
}

void Screen::init_colors ()
{
	TRF;

	switch (getInkset () )
	{
	case InkStandard:
		set_standard_default_inks ();
		break;
	case InkCpc:
		set_cpc_default_inks ();
		break;
	case InkSpectrum:
		set_spectrum_default_inks ();
		break;
	case InkMsx:
		set_msx_default_inks ();
		break;
	default:
		TRDEB ("Invalid inkset value");
		throw ErrBlassicInternal;
	}
	//colors_inited= true;
	//pforeground= mapcolor (default_pen).pc;
	//pbackground= mapcolor (default_paper).pc;

	raw_init_colors ();
}

void Screen::set_origin (int x, int y)
{
	originx= x;
	originy= y;
}

void Screen::set_limits (int minx, int maxx, int miny, int maxy)
{
	limited= true;

	sort_pair (minx, maxx);
	limit_minx= minx;
	limit_maxx= maxx;

	adjust_y (miny);
	adjust_y (maxy);
	sort_pair (miny, maxy);
	limit_miny= miny;
	limit_maxy= maxy;

	limit_minx= std::max (0, minx);
	limit_maxx= std::min (scwidth - 1, maxx);
	limit_miny= std::max (0, miny);
	limit_maxy= std::min (scheight - 1, maxy);

	if (limit_minx <= 0 && limit_maxx >= scwidth - 1 &&
		limit_miny <= 0 && limit_maxy >= scheight - 1)
	{
		limited= false;
	}
}

bool Screen::check_limit (int x, int y)
{
	if (x < 0 || y < 0 || x >= scwidth || y >= scheight)
		return false;
	return (! limited) || (x >= limit_minx && x <= limit_maxx &&
		y >= limit_miny && y <= limit_maxy);
}

void Screen::adjust_y (int & y) const
{
	switch (activetransform)
	{
	case TrIdentity:
		break; // Nothing to do
	case TrInvertY:
		y= scheight - 1 - y;
		break;
	}
}

void Screen::transform_x (int & x) const
{
	x+= originx;
}

void Screen::transform_y (int & y) const
{
	y+= originy;
	adjust_y (y);
}

void Screen::do_rotate (int & x, int & y) const
{
	real_rotate (x, y, rotatemode, scwidth);
}

void Screen::do_rotate (short int & x, short int & y) const
{
	real_rotate (x, y, rotatemode, scwidth);
}

void Screen::do_rotate_rel (int & x, int & y) const
{
	real_rotate_rel (x, y, rotatemode);
}

void Screen::do_rotate_rel (short int & x, short int & y) const
{
	real_rotate_rel (x, y, rotatemode);
}

void Screen::do_rotate_rel_abs (int & x, int & y) const
{
	real_rotate_rel_abs (x, y, rotatemode);
}

void Screen::do_rotate_rel_abs (short int & x, short int & y) const
{
	real_rotate_rel_abs (x, y, rotatemode);
}

void Screen::do_unrotate (int & x, int & y) const
{
	real_unrotate (x, y, rotatemode, scwidth);
}

void Screen::do_unrotate (short int & x, short int & y) const
{
	real_unrotate (x, y, rotatemode, scwidth);
}

void Screen::do_line_unmasked (int x, int y)
{
	int prevx= lastx, prevy= lasty;
	lastx= x; lasty= y;

	transform_x (x); transform_x (prevx);
	transform_y (y); transform_y (prevy);

	do_rotate (prevx, prevy);
	do_rotate (x, y);

	//setactivecolor (pforeground);
	//active_foreground ();
	setactiveink (graphics_pen);

	#if 1

	raw_draw_line (prevx, prevy, x, y);

	#else

	#ifdef BLASSIC_USE_X

	if (! fSynchro)
		XDrawLine (display, window, gc, prevx, prevy, x, y);
	XDrawLine (display, pixmap, gcp, prevx, prevy, x, y);
	XFlush (display);

	#elif defined BLASSIC_USE_WINDOWS_GRAPHICS

	if (! fSynchro)
	{
		MoveToEx (hdc, prevx, prevy, 0);
		LineTo (hdc, x, y);
		LineTo (hdc, x + 1, y); // Last point
	}

	MoveToEx (hdcPixmap, prevx, prevy, 0);
	LineTo (hdcPixmap, x, y);
	LineTo (hdcPixmap, x + 1, y); // Last point

	#endif

	#endif
}


#if 0
const unsigned char maskvaluedefault= '\xFF';
unsigned char maskvalue= maskvaluedefault;
bool maskdrawfirst= true;
unsigned maskpos= 0;
unsigned char auxmask []= { 1, 2, 4, 8, 16, 32, 64, 128 };
#endif

unsigned const char Screen::auxmask []= { 1, 2, 4, 8, 16, 32, 64, 128 };

void Screen::set_mask (unsigned char newmask)
{
	maskvalue= newmask;
	maskpos= 0;
}

void Screen::set_maskfirstpoint (bool firstpoint)
{
	maskdrawfirst= firstpoint;
}

void Screen::do_plot (int x, int y)
{
	if (! check_limit (x, y) )
		return;

	do_rotate (x, y);
	raw_plot (x, y);
}

void Screen::impl_plot_mask (int x, int y)
{
	if (maskvalue == maskvaluedefault)
		do_plot (x, y);
	else
	{
		if (maskvalue & auxmask [maskpos] )
			do_plot (x, y);
		if (++maskpos == 8) maskpos= 0;
	}
}

void Screen::do_line_mask (int x, int y)
{
	int prevx= lastx, prevy= lasty;
	lastx= x; lasty= y;

	transform_x (prevx); transform_x (x);
	transform_y (prevy); transform_y (y);

	//setactivecolor (pforeground);
	//active_foreground ();
	setactiveink (graphics_pen);

	int px= x - prevx;
	int py= y - prevy;
	int d1x= px < 0 ? -1 : px > 0 ? 1 : 0;
	int d1y= py < 0 ? -1 : py > 0 ? 1 : 0;
	int d2x= d1x;
	int d2y= 0;
	int m= abs (px);
	int n= abs (py);
	if (m <= n)
	{
		d2x= 0;
		d2y= d1y;
		m= abs (py);
		n= abs (px);
	}
	int s= m / 2;
	for (int i= 0; i <= m; ++i)
	{
		if (i != 0 || maskdrawfirst)
		{
			//if (maskvalue & auxmask [maskpos] )
			//	do_plot (prevx, prevy);
			//if (++maskpos == 8) maskpos= 0;
			impl_plot_mask (prevx, prevy);
		}
		s+= n;
		if (s >= m)
		{
			s-= m;
			prevx+= d1x;
			prevy+= d1y;
		}
		else
		{
			prevx+= d2x;
			prevy+= d2y;
		}
	}
}

void Screen::do_line (int x, int y)
{
	// When limited we use line masked to simplify line unmasked.
	if (! limited && (maskvalue == maskvaluedefault && maskdrawfirst) )
		do_line_unmasked (x, y);
	else
		do_line_mask (x, y);
}

void Screen::do_line_rel (int x, int y)
{
	do_line (lastx + x, lasty + y);
}

void Screen::do_move (int x, int y)
{
	lastx= x;
	lasty= y;
}

void Screen::do_move_rel (int x, int y)
{
	lastx+= x;
	lasty+= y;
}

#if 0
void Screen::do_scroll_up (int x, int y, int w, int h, int amount)
{
	TRACEFUNC (tr, "Screen::do_scroll_up");
	TRMESSAGE (tr, "x: " + to_string (x) + " y: " + to_string (y) +
		" w: " + to_string (w) + " h: " + to_string (h) +
		" amount: " + to_string (amount) );

	int x1= x;
	int x2= x1 + w;
	int y1= y;
	int y2= y1 + h;

	// Test
	int xt= x;
	int yt= y;
	int xf1= x;
	int yf1= y + amount;
	int xf2= x + w - 1;
	int yf2= yf1 + h - 1;

	// Reinit rectangle.
	int x1reinit= x1;
	int x2reinit= x1 + w;
	int y1reinit= y1;
	int y2reinit= y1 + h + amount;

	int y1fill= y1 + h;
	int y2fill= y1 + h + amount - 1;

	// Fill rectangle. Not rotated because
	// do_fill_rectangle will do it.
	int x1fill= x1;
	int x2fill= x1 + w - 1;

	do_rotate (x1, y1);
	do_rotate (x2, y2);
	sort_pair (x1, x2);
	sort_pair (y1, y2);

	do_rotate_rel_abs (w, h);

	int x1src= x1;
	int y1src= y1;
	switch (rotatemode)
	{
	case RotateNone:
		y1src+= amount;
		break;
	case Rotate90:
		x1src+= amount;
		break;
	}
	do_rotate (x1reinit, y1reinit);
	do_rotate (x2reinit, y2reinit);
	sort_pair (x1reinit, x2reinit);
	sort_pair (y1reinit, y2reinit);

	// Scrolling

	//raw_copy_rectangle (x1src, y1src, x1, y1, w, h);
	raw_copy_rectangle (xf1, yf1, xf2, yf2, xt, yt);

	// Fill new area with background color.

	//active_background ();

	setactiveink (graphics_paper);
	do_fill_rectangle (x1fill, y1fill, x2fill, y2fill, false);

	//active_foreground ();
}
#endif


void Screen::do_scroll (int x1, int y1, int x2, int y2, int xoff, int yoff,
	bool fill_it, InkNum fillink)
{
	TRF;

	do_rotate (x1, y1);
	do_rotate (x2, y2);
	sort_pair (x1, x2);
	sort_pair (y1, y2);
	do_rotate_rel (xoff, yoff);

	int copyx1= x1;
	int copyx2= x2;
	int destx1= x1;
	//int destx2= x2;
	if (xoff > 0)
	{
		destx1+= xoff;
		copyx2-= xoff;
	}
	else
	{
		//destx2+= xoff;
		copyx1-= xoff;
	}
	int copyy1= y1;
	int copyy2= y2;
	int desty1= y1;
	//int desty2= y2;
	if (yoff > 0)
	{
		desty1+= yoff;
		copyy2-= yoff;
	}
	else
	{
		//desty2+= yoff;
		copyy1-= yoff;
	}

	raw_copy_rectangle (copyx1, copyy1, copyx2, copyy2, destx1, desty1);

	if (fill_it)
	{
		setactiveink (fillink);
		if (yoff)
		{
			int yf1, yf2;
			if (yoff > 0)
			{
				yf1= y1;
				yf2= y1 + yoff - 1;
			}
			else
			{
				yf1= y2 + yoff + 1;
				yf2= y2;
			}
			direct_fill_rectangle (x1, yf1, x2, yf2);
		}
		if (xoff)
		{
			int xf1, xf2;
			if (xoff > 0)
			{
				xf1= x1;
				xf2= x1 + xoff - 1;
			}
			else
			{
				xf1= x2 + xoff + 1;
				xf2= x2;
			}
			direct_fill_rectangle (xf1, y1, xf2, y2);
		}
	}
}

void Screen::scroll (int x1, int y1, int x2, int y2, int xoff, int yoff,
	bool fill_it, InkNum fillink)
{
	TRF;

	transform_x (x1);
	transform_x (x2);
	transform_y (y1);
	transform_y (y2);
	do_scroll (x1, y1, x2, y2, xoff, yoff, fill_it, fillink);
}

void Screen::do_scroll_up (int x1, int y1, int x2, int y2,
	int amount, InkNum fillink)
{
	TRFDEBS ("x1: " << x1 << " y1: " << y1 <<
		" x2: " << x2 << " y2: " << y2 <<
		" amount: " << amount);

	#if 1
	do_scroll (x1, y1, x2, y2, 0, -amount, true, fillink);
	#else
	int xt= x1;
	int xt2= x2;
	int yt= y1;
	int yt2= y2;

	int xf1= x1;
	int yf1= y1 + amount;
	int xf2= x2;
	int yf2= y2;

	// Fill rectangle. Not rotated because
	// do_fill_rectangle will do it.
	int x1fill= x1;
	int x2fill= x2;
	int y1fill= y2 - amount + 1;
	int y2fill= y2;

	do_rotate (xf1, yf1);
	do_rotate (xf2, yf2);
	sort_pair (xf1, xf2);
	sort_pair (yf1, yf2);

	do_rotate (xt, yt);
	do_rotate (xt2, yt2);
	sort_pair (xt, xt2);
	sort_pair (yt, yt2);

	// Scrolling

	raw_copy_rectangle (xf1, yf1, xf2, yf2, xt, yt);

	// Fill new area with filling color.

	setactiveink (fillink);
	do_fill_rectangle (x1fill, y1fill, x2fill, y2fill, false);
	#endif
}

void Screen::do_scroll_down (int x1, int y1, int x2, int y2,
	int amount, InkNum fillink)
{
	TRFDEBS ("x1: " << x1 << " y1: " << y1 <<
		" x2: " << x2 << " y2: " << y2 <<
		" amount: " << amount);

	#if 1
	do_scroll (x1, y1, x2, y2, 0, amount, true, fillink);
	#else
	int xt= x1;
	int xt2= x2;
	int yt= y1 + amount;
	int yt2= y2;

	int xf1= x1;
	int yf1= y1;
	int xf2= x2;
	int yf2= y2 - amount;

	// Fill rectangle. Not rotated because
	// do_fill_rectangle will do it.
	int x1fill= x1;
	int x2fill= x2;
	int y1fill= y1;
	int y2fill= y1 + amount - 1;

	do_rotate (xf1, yf1);
	do_rotate (xf2, yf2);
	sort_pair (xf1, xf2);
	sort_pair (yf1, yf2);

	do_rotate (xt, yt);
	do_rotate (xt2, yt2);
	sort_pair (xt, xt2);
	sort_pair (yt, yt2);

	// Scrolling

	raw_copy_rectangle (xf1, yf1, xf2, yf2, xt, yt);

	// Fill new area with filling color.

	setactiveink (fillink);
	do_fill_rectangle (x1fill, y1fill, x2fill, y2fill, false);
	#endif
}

void Screen::do_scroll_left (int x1, int y1, int x2, int y2,
	int amount, InkNum fillink)
{
	TRFDEBS ("x1: " << x1 << " y1: " << y1 <<
		" x2: " << x2 << " y2: " << y2 <<
		" amount: " << amount);

	#if 1
	do_scroll (x1, y1, x2, y2, -amount, 0, true, fillink);
	#else
	int xt1= x1;
	int xt2= x1 + amount;
	int yt1= y1;
	int yt2= y2;

	int xf1= x1  + amount;
	int yf1= y1;
	int xf2= x2;
	int yf2= y2;

	int x1fill= x2 - amount + 1;
	int x2fill= x2;
	int y1fill= y1;
	int y2fill= y2;

	do_rotate (xf1, yf1);
	do_rotate (xf2, yf2);
	sort_pair (xf1, xf2);
	sort_pair (yf1, yf2);

	do_rotate (xt1, yt1);
	do_rotate (xt2, yt2);
	sort_pair (xt1, xt2);
	sort_pair (yt1, yt2);

	// Scrolling

	raw_copy_rectangle (xf1, yf1, xf2, yf2, xt1, yt1);

	// Fill new area with filling color.

	setactiveink (fillink);
	do_fill_rectangle (x1fill, y1fill, x2fill, y2fill, false);
	#endif
}

void Screen::invert_rectangle (int x1, int y1, int x2, int y2)
{
	do_rotate (x1, y1);
	do_rotate (x2, y2);
	if (x1 > x2)
		swap (x1, x2);
	if (y1 > y2)
		swap (y1, y2);
	raw_invert_rectangle (x1, y1, x2, y2);
}

void Screen::do_fill_rectangle (int x1, int y1, int x2, int y2, bool limitable)
{
	TRFDEBS ("x1: " << x1 << " x2: " << x2 <<
		" y1: " << y1 << " y2: " << y2);

	using std::min;
	using std::max;
	if (limitable && limited)
	{
		x1= max (x1, limit_minx);
		y1= max (y1, limit_miny);
		x2= min (x2, limit_maxx);
		y2= min (y2, limit_maxy);
		if (x1 > limit_maxx || x2 < limit_minx ||
				y1 > limit_maxy || y2 < limit_miny)
			return;
	}

	do_rotate (x1, y1);
	do_rotate (x2, y2);
	sort_pair (x1, x2);
	sort_pair (y1, y2);
	TRDEBS ("x1: " << x1 << " x2: " << x2 << " y1: " << y1 << " y2: " << y2);

	#if 0

	#ifdef BLASSIC_USE_WINDOWS_GRAPHICS

	RECT r = { x1, y1, x2 + 1, y2 + 1 };
	LOGPEN logpen;
	//GetObject (* pforeground, sizeof (LOGPEN), & logpen);
	GetObject (* activecolor, sizeof (LOGPEN), & logpen);
	HBRUSH hbrush= CreateSolidBrush (logpen.lopnColor);
	FillRect (hdcPixmap, & r, hbrush);
	if (! fSynchro)
		FillRect (hdc, & r, hbrush);
	DeleteObject (hbrush);

	#elif defined BLASSIC_USE_X

	//int w= std::abs (x2 - x1) + 1;
	//int h= std::abs (y2 - y1) + 1;
	int w= x2 - x1 + 1;
	int h= y2 - y1 + 1;
	XFillRectangle (display, pixmap, gcp,
		x1, y1, w, h);
	if (! fSynchro)
		XFillRectangle (display, window, gc,
			x1, y1, w, h);

	#endif

	#else

	direct_fill_rectangle (x1, y1, x2, y2);

	#endif
}

void Screen::cls ()
{
	TRF;

	setactiveink (graphics_paper);
	do_fill_rectangle (0, 0, get_width () - 1, get_height () - 1, true);
}

void Screen::rectangle (Point org, Point dest)
{
	int x1= org.x;
	int y1= org.y;
	int x2= dest.x;
	int y2= dest.y;
	lastx= x2; lasty= y2;

	transform_x (x1); transform_x (x2);
	transform_y (y1); transform_y (y2);

	do_rotate (x1, y1);
	do_rotate (x2, y2);

	if (x1 > x2) swap (x1, x2);
	if (y1 > y2) swap (y1, y2);

	setactiveink (graphics_pen);
	raw_rectangle (x1, y1, x2, y2);
}

void Screen::fill_rectangle (Point org, Point dest)
{
	int x1= org.x;
	int y1= org.y;
	int x2= dest.x;
	int y2= dest.y;
	lastx= x2; lasty= y2;

	transform_x (x1); transform_x (x2);
	transform_y (y1); transform_y (y2);

	do_rotate (x1, y1);
	do_rotate (x2, y2);

	sort_pair (x1, x2);
	sort_pair (y1, y2);

	using std::min;
	using std::max;
	if (limited)
	{
		x1= max (x1, limit_minx);
		y1= max (y1, limit_miny);
		x2= min (x2, limit_maxx);
		y2= min (y2, limit_maxy);
		if (x1 > limit_maxx || x2 < limit_minx ||
				y1 > limit_maxy || y2 < limit_miny)
			return;
	}

	setactiveink (graphics_pen);
	direct_fill_rectangle (x1, y1, x2, y2);
}

Inkset Screen::getInkset () const
{
	return inkset;
}

InkNum Screen::get_default_pen () const
{
	return default_pen;
}

InkNum Screen::get_default_paper () const
{
	return default_paper;
}

InkNum Screen::get_graphics_pen () const
{
	return graphics_pen;
}

InkNum Screen::get_graphics_paper () const
{
	return graphics_paper;
}

void Screen::set_graphics_pen (InkNum inknum)
{
	graphics_pen= inknum;
}

void Screen::set_graphics_paper (InkNum inknum)
{
	graphics_paper= inknum;
}

void Screen::set_transparent (bool active)
{
	transparent_text= active;
}

void Screen::plot (int x, int y)
{
	lastx= x; lasty= y;
	transform_x (x);
	transform_y (y);

	//active_foreground ();
	setactiveink (graphics_pen);

	do_plot (x, y);
}

void Screen::plot_rel (int x, int y)
{
	plot (lastx + x, lasty + y);
}

void Screen::zx (Point p)
{
	p.x= (p.x / 2 + 1) * 4;
	p.y= (21 - p.y / 2) * 4;
	Point p2 (p);
	p2.x+= 3;
	p2.y+= 3;
	fill_rectangle (p, p2);
}

void Screen::zx_plot (Point p)
{
	setactiveink (graphics_pen);
	zx (p);
}

void Screen::zx_unplot (Point p)
{
	setactiveink (graphics_paper);
	zx (p);
}

int Screen::test (int x, int y, bool relative)
{
	if (relative)
	{
		lastx+= x;
		lasty+= y;
	}
	else
	{
		lastx= x;
		lasty= y;
	}
	x= lastx;
	y= lasty;
	transform_x (x);
	transform_y (y);

	if (x < 0 || y < 0 || x >= scwidth || y >= scheight)
		return 0;

	do_rotate (x, y);
	return raw_test (x, y);
}

inline void rotate_char_image (unsigned char (& ch) [8], RotateMode rotmode)
{
	switch (rotmode)
	{
	case RotateNone:
		break;
	case Rotate90:
		{
			unsigned char ch2 [8];
			unsigned char maski= 0x80;
			for (size_t i= 0; i < 8; ++i)
			{
				unsigned char maskj= 0x01;
				unsigned char c= 0;
				for (size_t j= 0; j < 8; ++j)
				{
					if (ch [j] & maski)
						c|= maskj;
					maskj<<= 1;
				}
				ch2 [i]= c;
				maski>>= 1;
			}
			for (size_t i= 0; i < 8; ++i)
				ch [i]= ch2 [i];
		}
		break;
	}
}

std::string Screen::copychrat (int x, int y, BlChar from, BlChar to)
{
	TRFDEBS ("x: " << x << " y: " << y);

	if (x < 0 || x > scwidth - charwidth ||
		y < 0 || y > scheight - charheight)
	{
		TRDEB ("Out of range");
		return std::string ();
	}

	int xzoom= zoomtextx;
	int yzoom= zoomtexty;
	do_rotate_rel_abs (xzoom, yzoom);

	int x1= x, x2= x + charwidth - 1;
	int y1= y, y2= y + charheight - 1;
	do_rotate (x1, y1);
	do_rotate (x2, y2);
	sort_pair (x1, x2);
	sort_pair (y1, y2);

	unsigned char ch [8]= {0};
	if (! load_char_image (x1, y1, xzoom, yzoom, ch) )
	{
		TRDEB ("No char image");
		return std::string ();
	}

	rotate_char_image (ch, rotatemode);

	unsigned char chinv [8];
	for (int i= 0; i < 8; ++i)
		chinv [i]= static_cast <unsigned char> (~ ch [i]);

	int iFrom= static_cast <int> (from);
	int iTo= static_cast <int> (to) + 1;
	for (int i= iFrom; i < iTo; ++i)
	{
		if (memcmp (charset::data [i], ch, 8) == 0 ||
			memcmp (charset::data [i], chinv, 8) == 0)
		{
			return std::string (1, static_cast <char> (i) );
		}
	}

	return std::string ();
}

std::string Screen::copychratpos (int col, int row, BlChar from, BlChar to)
{
	return copychrat (col * charwidth, row * charheight, from, to);
}

Image * Screen::getImage (int x1, int y1, int x2, int y2)
{
	transform_x (x1);
	transform_y (y1);
	transform_x (x2);
	transform_y (y2);

	if (x2 < 0 || y2 < 0 || x1 >= scwidth || y1 >= scheight)
		throw ErrFunctionCall;
	if (x1 < 0)
		x1= 0;
	if (y1 < 0)
		y1= 0;
	if (x2 >= scwidth)
		x2= scwidth - 1;
	if (y2 >= scheight)
		y2= scheight - 1;

	do_rotate (x1, y1);
	do_rotate (x2, y2);
	sort_pair (x1, x2);
	sort_pair (y1, y2);

	return raw_getImage (x1, y1, x2, y2);
}

void Screen::putImage (Image & image, int x, int y, int mode)
{
	transform_x (x);
	transform_y (y);
	if (x >= scwidth || y >= scheight)
		throw ErrFunctionCall;
	int w= image.getwidth ();
	int h= image.getheight ();
	do_rotate_rel_abs (w, h);
	int x2= x + w - 1;
	int y2= y + h - 1;
	do_rotate (x, y);
	do_rotate (x2, y2);
	sort_pair (x, x2);
	sort_pair (y, y2);
	raw_putImage (image, x, y, mode);
}

void Screen::paint (InkNum inkpaint, InkNum inkborder, int x, int y)
{
	TRF;

	auto_ptr <Painter> ap;

	int minx= limit_minx;
	int miny= limit_miny;
	int maxx= limit_maxx;
	int maxy= limit_maxy;
	do_rotate (minx, miny);
	do_rotate (maxx, maxy);
	sort_pair (minx, maxx);
	sort_pair (miny, maxy);

	Painter * p= raw_getPainter (inkpaint, inkborder,
		minx, miny, maxx, maxy);
	ap.reset (p);

	transform_x (x);
	transform_y (y);
	do_rotate (x, y);

	//setactiveink (inkpaint);
	p->paint (x, y);

	//delete p;
}

void Screen::do_plot_points (ShortPoint point [], int npoints, bool limitable)
{
	TRF;

	ASSERT (npoints < static_cast <int> (64 * zoomtexty) );

	if (rotatemode != RotateNone)
		for (int i= 0; i < npoints; ++i)
			do_rotate (point [i].x, point [i].y);

	if (zoomtextx == 1 && zoomtexty == 1)
	{
		real_plot_points (point, npoints);
	}
	else
	{
		int xinc= zoomtextx, yinc= zoomtexty;
		do_rotate_rel (xinc, yinc);
		if (yinc < 0)
		{
			yinc= -yinc;
			if (yinc > 1)
			{
				short int des= static_cast <short int>
					(yinc - 1);
				for (int i= 0; i < npoints; ++i)
					point [i].y-= des;
			}
		}

		int xmin= 0, xmax= realwidth;
		int ymin= 0, ymax= realheight;
		if (limitable)
		{
			xmin= limit_minx;
			xmax= limit_maxx;
			ymin= limit_miny;
			ymax= limit_maxy;
			do_rotate (xmin, ymin);
			do_rotate (xmax, ymax);
			sort_pair (xmin, xmax);
			sort_pair (ymin, ymax);
		}

		real_plot_wide_points (point, npoints, xinc, yinc,
			xmin, ymin, xmax, ymax);
	}
}


// --------------------------------
// drawarc auxiliary functions.
// --------------------------------

// Workaround for a problem on gcc.
double zero= 0.0; // Don't make this var const!

inline int signum (double d)
{
	return d < zero ? -1 : d > zero ? 1 : 0;
}

inline int roundint (double d)
{
	double frac= modf (d, & d);
	if (frac > 0.5)
		++d;
	else
		if (frac < -0.5)
			--d;
	return int (d);
}


void Screen::do_drawarcsegment (int xd, int yd,
	int & coords_x, int & coords_y, bool & firstpoint)
{
	// This function is used instead of the other line
	// draw functions to ensure exact compatibility with
	// the Spectrum and because of first point of line
	// treatement.

	int b= abs (yd);
	int c= abs (xd);
	int d= signum (yd);
	int e= signum (xd);
	//std::cerr << "do_drawarcsegment " << c << ' ' << b << ' ' <<
	//	e << ' ' << d << std::endl;
	int h, l, hvx, hvy;
	if (c >= b)
	{
		h= c; l= b; hvx= e; hvy= 0;
	}
	else
	{
		if (b == 0)
			return;
		h= b; l= c; hvx= 0; hvy= d;
	}
	b= h;
	int a= h / 2;
	for (int i= 0; i < b; ++i)
	{
		a+= l;
		if (a < h)
		{
			// Horizontal or vertical step.
			coords_x+= hvx;
			coords_y+= hvy;
		}
		else
		{
			// Diagonal step.
			a-= h;
			coords_x+= e;
			coords_y+= d;
		}
		if (firstpoint)
			firstpoint= false;
		else
		{
			int x= coords_x; int y= coords_y;
			transform_x (x);
			transform_y (y);
			impl_plot_mask (x, y);
		}
	}
}

void Screen::do_drawarc (int x, int y, double g)
{
	// Mimic the Spectrum DRAW instruction, even with values
	// of angle greater than 2 * pi.
	// The original algorithm uses the Spectrum cooordinates,
	// then the angle is inverted when using non inverted
	// coordinates.
	// To an explanation of the algorithm used, see "The Complete
	// Spectrum ROM Disassembly" or some other disassembly of the
	// Spectrum ROM (the book has more complete comments).

	switch (activetransform)
	{
	case TrIdentity:
		g= -g;
		break;
	case TrInvertY:
		break; // Nothing to do
	}

	//setactivecolor (pforeground);
	//active_foreground ();
	setactiveink (graphics_pen);

	// If drawing of first point is active, no special treatement
	// for the first point by marking as if not were the first.
	bool firstpoint= ! maskdrawfirst;

	int xend= lastx + x, yend= lasty + y;

	double sing2= sin (g / 2);
	double z;
	if (sing2 != 0 && (z= fabs ( (abs (x) + abs (y) ) / sing2) ) >= 1)
	{
		int a= roundint (fabs (g * sqrt (z) / 2) );
		if (a <= 255)
		{
			a= (a & 0xFC) + 4;
			if (a >= 256)
				a= 252;
		}
		else
			a= 252;

		double singa= sin (g / a);
		double cosga= cos (g / a);
		double w= sin (g / (2.0 * a) ) / sing2;
		double f= g / 2 - g / (2.0 * a);
		double sinf= sin (f);
		double cosf= cos (f);
		double un= y * w * sinf + x * w * cosf;
		double vn= y * w * cosf - x * w * sinf;

		double uv= fabs (un) + fabs (vn);
		if (uv >= 1)
		{
			double xn= lastx;
			double yn= lasty;
			a= a - 1;
			while (a > 0)
			{
				xn= xn + un;
				yn= yn + vn;
				do_drawarcsegment (
					roundint (xn - lastx),
					roundint (yn - lasty),
					lastx, lasty, firstpoint);
				a= a - 1;
				if (a > 0)
				{
					double un1= un;
					un= un1 * cosga - vn * singa;
					vn= un1 * singa + vn * cosga;
				}
			}
		}
	}
	// Draw the last segment (can be the entire arc).
	do_drawarcsegment (xend - lastx, yend - lasty,
		lastx, lasty, firstpoint);
}

//#define DEBUG_CIRCLE

#ifdef DEBUG_CIRCLE

#include <unistd.h>

namespace { inline void pausec () { usleep (100000); } }

#else

namespace { inline void pausec () { } }

#endif

void Screen::circle (int x, int y, int radius)
{
	TRF;

	//active_foreground ();
	setactiveink (graphics_pen);

	lastx= x + radius;
	lasty= y;

	transform_x (x);
	transform_y (y);

	if (radius == 0)
	{
		impl_plot_mask (x, y);
		return;
	}

	// sq2_2 is sin (pi/4)
	#ifdef M_SQRT2
	static const double sq2_2= M_SQRT2 / 2.0;
	#else
	static const double sq2_2= sqrt (2.0) / 2.0;
	#endif
	//int r= int (radius * sq2_2 + .5) + 1;
	int r= int (radius * sq2_2 + 1.0);
	int rr= int (sqrt (radius * radius - r * r) + .5);
	int dim= r;
	if (rr >= r) ++dim;

	#ifdef DEBUG_CIRCLE
	cerr << "Circle: " << radius << ", " << r << endl;
	#endif

	util::auto_buffer <int> p (dim);

	// Bresenham algorithm.
	for (int i= 0, j= radius, d= 1 - radius; i < dim; ++i)
	{
		p [i]= j;
		if (d < 0)
			d+= 2 * i + 3;
		else
		{
			d+= 2 * (i - j) + 5;
			--j;
		}
	}

	rr= p [r - 1] - 1;
	ASSERT (rr <= dim - 1);

	// The first point in each quadrant is plotted independently.
	// In the first quadrant is omitted, we plot it at the end.
	#ifdef DEBUG_CIRCLE
	blassic::graphics::setcolor (4);
	#endif
	for (int j= 1; j < r; ++j)
	{
		//do_line (x + p [j], y - j);
		impl_plot_mask (x + p [j], y - j);
		pausec ();
	}
	#ifdef DEBUG_CIRCLE
	blassic::graphics::setcolor (0);
	#endif
	for (int i= rr; i > 0; --i)
	{
		//do_line (x + i, y - p [i] );
		impl_plot_mask (x + i, y - p [i] );
		pausec ();
	}

	//do_line (x, y - radius);

	#ifdef DEBUG_CIRCLE
	blassic::graphics::setcolor (4);
	#endif

	impl_plot_mask (x, y - radius);
	for (int i= 1; i < r; ++i)
	{
		//do_line (x - i, y - p [i] );
		impl_plot_mask (x - i, y - p [i] );
		pausec ();
	}

	#ifdef DEBUG_CIRCLE
	blassic::graphics::setcolor (0);
	#endif

	for (int j= rr; j > 0; --j)
	{
		//do_line (x - p [j], y - j);
		impl_plot_mask (x - p [j], y - j);
		pausec ();
	}

	//do_line (x - radius, y);

	#ifdef DEBUG_CIRCLE
	blassic::graphics::setcolor (4);
	#endif

	impl_plot_mask (x - radius, y);
	for (int j= 1; j < r; ++j)
	{
		//do_line (x - p [j], y + j);
		impl_plot_mask (x - p [j], y + j);
		pausec ();
	}
	#ifdef DEBUG_CIRCLE
	blassic::graphics::setcolor (0);
	#endif
	for (int i= rr; i > 0; --i)
	{
		//do_line (x - i, y + p [i] );
		impl_plot_mask (x - i, y + p [i] );
		pausec ();
	}

	//do_line (x, y + radius);
	impl_plot_mask (x, y + radius);

	#ifdef DEBUG_CIRCLE
	blassic::graphics::setcolor (4);
	#endif

	for (int i= 1; i < r; ++i)
	{
		//do_line (x + i, y + p [i] );
		impl_plot_mask (x + i, y + p [i] );
		pausec ();
	}

	#ifdef DEBUG_CIRCLE
	blassic::graphics::setcolor (0);
	#endif

	for (int j= rr; j > 0; --j)
	{
		//do_line (x + p [j], y + j);
		impl_plot_mask (x + p [j], y + j);
		pausec ();
	}
	//do_line (x + radius, y);

	#ifdef DEBUG_CIRCLE
	blassic::graphics::setcolor (4);
	#endif

	impl_plot_mask (x + radius, y);

}


inline void get_point_on_arc (int r, BlNumber a, int & out_x, int & out_y)
{
	BlNumber s= sin (a) * r, c= cos (a) * r;
	out_x= static_cast <int> (c < 0 ? c - .5 : c + .5);
	out_y= static_cast <int> (s < 0 ? s - .5 : s + .5);
}

inline int get_quadrant (int x, int y)
{
	if (x >= 0)
	{
		if (y >= 0)
			return 0;
		else
			return 3;
	}
	else
	{
		if (y > 0)
			return 1;
		else
			return 2;
	}
}


void Screen::arccircle (int x, int y, int radius,
	BlNumber arcbeg, BlNumber arcend)
{
	TRF;

	/*
		The code for this function and his auxiliarys
		is taken from the Allegro library.
		Many thanks.
	*/

	//active_foreground ();
	setactiveink (graphics_pen);

	//#define DEBUG_ARCCIRCLE

	#ifdef DEBUG_ARCCIRCLE
	cerr << "arccircle: " << x << ", " << y << ", " << radius << ", " <<
		arcbeg << ", " << arcend << endl;
	#endif

	lastx= x + radius;
	lasty= y;
	transform_x (x);
	transform_y (y);

	int px, py; // Current position and auxiliary.
	get_point_on_arc (radius, arcend, px, py);
	const int ex= px, ey= py; // End position.
	get_point_on_arc (radius, arcbeg, px, py);
	const int sx= px, sy= py; // Start position.
	// Current position start at start position.

	const int sq= get_quadrant (sx, sy); // Start quadrant.
	// Calculate end quadrant, considering that end point
	// must be after start point.
	int q= get_quadrant (ex, ey);
	if (sq > q)
		// Quadrant end must be greater or equal.
		q+= 4;
	else if (sq == q && arcbeg > arcend)
		// If equal, consider the angle.
		q+= 4;
	const int qe= q;
	q= sq; // Current cuadrant.

	#ifdef DEBUG_ARCCIRCLE
	cerr << "Quadrant from " << sq << " to " << qe << endl;
	#endif

	// Direction of movement.
	int dy = ( ( (q + 1) & 2) == 0) ? 1 : -1;
	int dx= ( (q & 2) == 0) ? -1 : 1;

	const int rr= radius * radius;
	int xx= px * px;
	int yy= py * py - rr;

	while (true)
	{
		// Change quadrant when needed, adjusting directions.
		if ( (q & 1) == 0)
		{
			if (px == 0)
			{
				if (qe == q)
					break;
				++q;
				dy= -dy;
			}
		}
		else
		{
			if (py == 0)
			{
				if (qe == q)
					break;
				++q;
				dx= -dx;
			}
		}
		// If we are in the end quadrant, check if at the end position.
		if (qe == q)
		{
			int det= 0;
			if (dy > 0)
			{
				if (py >= ey)
					++det;
			}
			else
			{
				if (py <= ey)
					++det;
			}
			if (dx > 0)
			{
				if (px >= ex)
					++det;
			}
			else
			{
				if (px <= ex)
					++det;
			}
			if (det == 2)
				break;
		}

		impl_plot_mask (x + px, y - py);

		int xx_new= (px + dx) * (px + dx);
		int yy_new= (py + dy) * (py + dy) - rr;
		int rr1= xx_new + yy;
		int rr2= xx_new + yy_new;
		int rr3= xx + yy_new;
		if (rr1 < 0) rr1= -rr1;
		if (rr2 < 0) rr2= -rr2;
		if (rr3 < 0) rr3= -rr3;
		if (rr3 >= std::min (rr1, rr2) )
		{
			px+= dx;
			xx= xx_new;
		}
		if (rr1 > std::min (rr2, rr3) )
		{
			py+= dy;
			yy= yy_new;
		}
	}
	if (px != sx || py != sy || sq == qe)
		impl_plot_mask (x + px, y - py);
}

void Screen::ellipse (int ox, int oy, int rx, int ry)
{
	TRF;

	// Based on "A fast Bresenham type algorithm
	// for drawing ellipses", by John Kennedy.

	//#define DEBUG_ELLIPSE

	#ifdef DEBUG_ELLIPSE
	cerr << "Ellipse: " << rx << ", " << ry << endl;
	#endif

	//active_foreground ();
	setactiveink (graphics_pen);

	lastx= ox+ rx;
	lasty= oy;

	transform_x (ox);
	transform_y (oy);

	const int ry2= ry * ry;
	const int rx2= rx * rx;
	//const int dimy= int (ry2 / sqrt (ry2 + rx2) + 0.5);
	const int dimy= int (ry2 / sqrt (ry2 + rx2) ) + 1;

	#ifdef DEBUG_ELLIPSE
	cerr << "Dimy: " << dimy << endl;
	#endif

	util::auto_buffer <int> py (dimy);
	//#define SIMPLER
	#ifndef SIMPLER
	const int ry2_2= ry2 * 2;
	const int rx2_2= rx2 * 2;
	int xchange= ry2 * (1 - 2 * rx);
	int ychange= rx2;
	int ellipseerror= 0;
	for (int y= 0, x= rx; y < dimy; ++y)
	{
		py [y]= x;
		if (x == 0) // Needed for little rx
			continue;
		ellipseerror+= ychange;
		ychange+= rx2_2;
		if (2 * ellipseerror + xchange > 0)
		{
			--x;
			ellipseerror+= xchange;
			xchange+= ry2_2;
		}
	}
	#else
	for (int y= 0; y < dimy; ++y)
		py [y]= int (rx * sqrt (ry * ry - y * y) / ry + 0.5);
	#endif

	int aux= dimy > 0 ? py [dimy - 1] : rx;
	const int dimx= aux < 0 ? 0 : aux;

	#ifdef DEBUG_ELLIPSE
	cerr << "Dimx: " << dimx << endl;
	#endif

	util::auto_buffer <int> px (dimx);
	#ifndef SIMPLER
	xchange= ry2;
	ychange= rx2 * (1 - 2 * ry);
	ellipseerror= 0;
	for (int x= 0, y= ry; x < dimx; ++x)
	{
		px [x]= y;
		if (y == 0) // Needed for little ry
			continue;
		ellipseerror+= xchange;
		xchange+= ry2_2;
		if (2 * ellipseerror + ychange > 0)
		{
			--y;
			ellipseerror+= ychange;
			ychange+= rx2_2;
		}
	}
	#else
	for (int x= 0; x < dimx; ++x)
		px [x]= int (ry * sqrt (rx * rx - x * x) / rx + 0.5);
	#endif

	for (int y= 1; y < dimy; ++y)
		impl_plot_mask (ox + py [y], oy - y);
	for (int x= dimx - 1; x > 0; --x)
		impl_plot_mask (ox + x, oy - px [x] );
	impl_plot_mask (ox, oy - ry);
	for (int x= 1; x < dimx; ++x)
		impl_plot_mask (ox - x, oy - px [x] );
	for (int y= dimy - 1; y > 0; --y)
		impl_plot_mask (ox - py [y], oy - y);
	impl_plot_mask (ox - rx, oy);
	for (int y= 1; y < dimy; ++y)
		impl_plot_mask (ox - py [y], oy + y);
	for (int x= dimx - 1; x > 0; --x)
		impl_plot_mask (ox - x, oy + px [x] );
	impl_plot_mask (ox, oy + ry);
	for (int x= 1; x < dimx; ++x)
		impl_plot_mask (ox + x, oy + px [x] );
	for (int y= dimy - 1; y > 0; --y)
		impl_plot_mask (ox + py [y], oy + y);
	impl_plot_mask (ox + rx, oy);
}


inline void get_point_on_arc (int rx, int ry, BlNumber a,
	int & out_x, int & out_y)
{
	BlNumber s= sin (a) * ry, c= cos (a) * rx;
	out_x= static_cast <int> (c < 0 ? c - .5 : c + .5);
	out_y= static_cast <int> (s < 0 ? s - .5 : s + .5);
}


void Screen::arcellipse (int ox, int oy, int rx, int ry,
	BlNumber arcbeg, BlNumber arcend)
{
	TRF;

	/*
		Adapted from the arccircle algorithm,
		not optimal but works.
	*/
	//#define DEBUG_ARCELLIPSE

	//active_foreground ();
	setactiveink (graphics_pen);

	lastx= ox+ rx;
	lasty= oy;

	transform_x (ox);
	transform_y (oy);

	int px, py; // Current position and auxiliary
	get_point_on_arc (rx, ry, arcend, px, py);
	const int ex= px, ey= py; // End position
	get_point_on_arc (rx, ry, arcbeg, px, py);
	const int sx= px, sy= py; // Start position

	const int sq= get_quadrant (sx, sy); // Start quadrant.
	// Calculate end quadrant, considering that end point
	// must be after start point.
	int q= get_quadrant (ex, ey);
	if (sq > q)
		// Quadrant end must be greater or equal.
		q+= 4;
	else if (sq == q && arcbeg > arcend)
		// If equal, consider the angle.
		q+= 4;
	const int qe= q;
	q= sq; // Current cuadrant.

	#ifdef DEBUG_ARCELLIPSE
	cerr << "Arc llipse from: " << arcbeg << " to " << arcend << endl;
	cerr << "Quadrant from " << sq << " to " << qe << endl;
	#endif

	// Direction of movement.
	int dy = ( ( (q + 1) & 2) == 0) ? 1 : -1;
	int dx= ( (q & 2) == 0) ? -1 : 1;

	const int rx2= rx * rx;
	const int ry2= ry * ry;
	const int rxy2= rx2 * ry2;

	while (true)
	{
		// Change quadrant when needed, adjusting directions.
		if ( (q & 1) == 0)
		{
			// Take care that in very eccentric ellipses
			// can be in 0 before the extreme point.
			if (px == 0 && abs (py) == ry)
			{
				if (qe == q)
					break;
				++q;
				dy= -dy;
			}
		}
		else
		{
			if (py == 0 && abs (px) == rx)
			{
				if (qe == q)
					break;
				++q;
				dx= -dx;
			}
		}
		// If we are in the end quadrant, check if at the end position.
		if (qe == q)
		{
			int det= 0;
			if (dy > 0)
			{
				if (py >= ey)
					++det;
			}
			else
			{
				if (py <= ey)
					++det;
			}
			if (dx > 0)
			{
				if (px >= ex)
					++det;
			}
			else
			{
				if (px <= ex)
					++det;
			}
			if (det == 2)
				break;
		}

		impl_plot_mask (ox + px, oy - py);

		int rr1= ry2 * (px + dx) * (px + dx) +
			rx2 * py * py - rxy2;
		int rr2= ry2 * (px + dx) * (px + dx) +
			rx2 * (py + dy) * (py + dy) - rxy2;
		int rr3= ry2 * px * px +
			rx2 * (py + dy) * (py + dy) - rxy2;
		if (rr1 < 0) rr1= -rr1;
		if (rr2 < 0) rr2= -rr2;
		if (rr3 < 0) rr3= -rr3;
		if (rr3 >= std::min (rr1, rr2) )
		{
			px+= dx;
			//xx= xx_new;
		}
		if (rr1 > std::min (rr2, rr3) )
		{
			py+= dy;
			//yy= yy_new;
		}
	}
	if (px != sx || py != sy || sq == qe)
		impl_plot_mask (ox + px, oy - py);
}

void Screen::do_printxy (unsigned char ch, int x, int y,
	InkNum chpen, InkNum chpaper,
	bool limitable, bool inverse, bool underline)
{
	TRF;

	static unsigned char mask [8]= { 128, 64, 32, 16, 8, 4, 2, 1 };

	TRDEB ("Preparing char");

	//charset::chardata & data= charset::data [ch];

	charset::chardata data;
	memcpy (data, charset::data [ch], sizeof (charset::chardata) );
	if (underline)
	{
		TRDEB ("underlined");
		data [7]= 255;
	}

	ShortPoint point [64]; // 64 pixels

	const int xzoom= zoomtextx;
	const int yzoom= zoomtexty;

	TRDEBS ("xzoom: " << xzoom << " yzoom: " << yzoom);
	ASSERT (xzoom != 0);
	ASSERT (yzoom != 0);

	TRDEB ("Cheking limits");

	int xmin= x;
	int xmax= x + 8 * xzoom - 1;
	int ymin= y;
	int ymax= y + 8 * yzoom - 1;
	if (limitable)
	{
		xmin= std::max (xmin, limit_minx);
		xmax= std::min (xmax, limit_maxx);
		ymin= std::max (ymin, limit_miny);
		ymax= std::min (ymax, limit_maxy);
	}

	int jmin= (xmin - x) / xzoom;
	int xbeg= x + jmin * xzoom;
	int imin= (ymin - y) / yzoom;
	int ybeg= y + imin * yzoom;

	TRDEB ("Writing char");

	int n= 0;
	for (int i= imin, yi= ybeg; yi <= ymax; ++i, yi+= zoomtexty)
	{
		unsigned char c= data [i];
		if (inverse)
			c= static_cast <unsigned char> (~ c);

		// Little optimization:
		if (c == 0)
			continue;

		for (int j= jmin, xj= xbeg; xj <= xmax; ++j, xj+= zoomtextx)
		{
			if (c & mask [j] )
			{
				point [n].x= static_cast <short> (xj);
				point [n].y= static_cast <short> (yi);
				++n;
			}
		}
	}
	if (n < 64)
	{
		if (! transparent_text)
		{
			//active_background ();
			setactiveink (chpaper);
			do_fill_rectangle (x, y,
				x + charwidth - 1, y + charheight - 1,
				limitable);
		}
		if (n > 0)
		{
			//active_foreground ();
			setactiveink (chpen);
			do_plot_points (point, n, limitable);
		}
	}
	else
	{
		setactiveink (chpen);
		do_fill_rectangle (x, y,
			x + charwidth - 1, y + charheight - 1,
			limitable);
	}
}

void Screen::printxy (unsigned char ch,
	bool limitable, bool inverse, bool underline)
{
	TRF;

	int x= lastx;
	int y= lasty;
	lastx+= charwidth;

	transform_x (x);
	transform_y (y);
	do_printxy (ch, x, y, graphics_pen, graphics_paper,
		limitable, inverse, underline);
}

void Screen::print (unsigned char ch, int col, int row,
	InkNum chpen, InkNum chpaper,
	bool inverse, bool underline)
{
	TRF;

	do_printxy (ch, col * charwidth, row * charheight,
		chpen, chpaper,
		false, inverse, underline);
}

std::string Screen::inkey ()
{
	QueueKey & queuekey= getQueueKey ();
	if (queuekey.empty () )
		return std::string ();
	else
		return queuekey.pop ();
}

std::string Screen::getkey ()
{
	QueueKey & queuekey= getQueueKey ();
	while (queuekey.empty () )
	{
		wait_event ();
		if (fInterrupted)
			return std::string ();
	}
	return queuekey.pop ();
}

int Screen::get_xmouse () const
{
	int x= raw_xmouse ();
	int y= raw_ymouse ();
	do_unrotate (x, y);
	return x;
}

int Screen::get_ymouse () const
{
	int x= raw_xmouse ();
	int y= raw_ymouse ();
	do_unrotate (x, y);
	return y;
}


#if 0
void Screen::check_events ()
{
	// Do nothing
}
#endif


} // namespace graphics_impl


#endif


} // namespace blassic



// End of graphics_impl.cpp
